/*
  Jind - an interpreter for a toy programming language
  Copyright (C) 2000, 2001 Markus Triska

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA

  Contact me at triska@gmx.at
*/


import java.util.Vector;

public class Game
{
    public Field field;
    public boolean won = false;
    public Player player = new Player(1, 1);

    /* Change these definitions to translate Jind to your language */

    final String tok_left = "left";
    final String tok_right = "right";
    final String tok_ahead = "ahead";
    final String tok_behind = "behind";

    final String tok_block = "block";
    final String tok_free = "free";
    final String tok_target = "target";

    final String tok_if = "if";
    final String tok_else = "else";
    final String tok_endif = "endif";
    final String tok_until = "until";
    final String tok_enduntil = "enduntil";

    final String tok_goahead = "goahead";
    final String tok_turnright = "turnright";
    final String tok_turnleft = "turnleft";
    final String tok_turnaround = "turnaround";

    String errmsgs[] = new String[6];

    /*  ========================================     */

    final int AHEAD  = 1;
    final int RIGHT  = 2;
    final int BEHIND = 3;
    final int LEFT   = 4;

    final int ERR_NONE = -1;
    final int ERR_SYNTAX = 0;
    final int ERR_ENDIF = 1;
    final int ERR_ENDUNTIL = 2;
    final int ERR_EOF = 3;
    final int ERR_BUMP = 4;
    final int ERR_DEADCYCLES = 5;


    int errnum = -1;
    int deadcycles = 0;
    final int maxdeadcycles = 10;


    Vector commands = new Vector();
    Vector loops = new Vector();
    int currcommand = -1;


    public Game(String map, String instrs)
    {
	commands.clear();
	String[] cs = instrs.split("\n");
	for (int i = 0; i < cs.length; i++) {
	    commands.add(cs[i].trim().toLowerCase());
	}
	errmsgs[ERR_SYNTAX] = "Syntax error on line ";
	errmsgs[ERR_ENDIF] = "Missing ENDIF";
	errmsgs[ERR_ENDUNTIL] = "Missing ENDUNTIL";
	errmsgs[ERR_EOF] = "No more instructions";
	errmsgs[ERR_BUMP] = "Game Over";
	errmsgs[ERR_DEADCYCLES] = "Too many dead cycles";

	field = new Field(map);
    }


    boolean contains(String s1, String s2)
    {
	return (s1.indexOf(s2) != -1);
    }

    byte what(String string)
    {
	byte w = 0;
	if (contains(string, tok_block)) {
	    w = Globals.BLOCK;
	} else if (contains(string, tok_target)) {
	    w = Globals.END;
	} else if (contains(string, tok_free)) {
	    w = Globals.FREE;
	}
	return w;
    }

    int direction(String string)
    {
	int d = 0;
	if (contains(string, tok_ahead)) {
	    d = AHEAD;
	} else if (contains(string, tok_right)) {
	    d = RIGHT;
	} else if (contains(string, tok_behind)) {
	    d = BEHIND;
	} else if (contains(string, tok_left)) {
	    d = LEFT;
	}
	return d;
    }

    boolean find_statement(String stmt)
    {
	while(true) {
	    currcommand += 1;
	    if (currcommand >= commands.size()) return false;

	    String command = (String) commands.elementAt(currcommand);

	    // Return when searching for ELSE, but none specified
	    if (stmt.equals(tok_else) && command.equals(tok_endif))
		return true;

	    // Ignore nested IFs when searching for ELSE
	    if (stmt.equals(tok_else) && command.equals(tok_if))
		find_statement(tok_endif);

	    // Ignore nested UNTILs when searching for ENDUNTIL
	    if (stmt.equals(tok_enduntil) && command.equals(tok_until))
		find_statement(tok_enduntil);

	    if (stmt.equals(command)) return true;
	}
    }

    public boolean test(int d, byte w)
    {
	if (w == Globals.END) {
	    if ((player.position().x == 49) && (player.position().y == 19))  {
		won = true;
		return true;
	    }
	}

	Position pos = new Position(0, 0);
	int x = player.position().x;
	int y = player.position().y;

	int ld = d + player.direction() - 1;
	if (ld > 4) ld -= 4;

	switch (ld) {
	case AHEAD:  pos = new Position(x, y - 1); break;
	case RIGHT:  pos = new Position(x + 1, y); break;
	case BEHIND: pos = new Position(x, y + 1); break;
	case LEFT:   pos = new Position(x - 1, y); break;
	}
	return field.test(pos, w);
    }

    public boolean isError() { return (errnum != ERR_NONE); }

    public String getErrorMessage() {
	if (errnum == ERR_SYNTAX) {
	    return errmsgs[ERR_SYNTAX] + (currcommand + 1) + ".";
	}
	return errmsgs[errnum];
    }

    public boolean next_statement()
    {
	/* very simple interpreter operating directly on the source;
	   simple pattern/string tests are used instead of actual parsing */

	currcommand += 1;
	if (currcommand >= commands.size()) {
	    errnum = ERR_EOF;
	    return false;
	}

	String command = (String) commands.elementAt(currcommand);

	if (command.length() == 0 || command.charAt(0) == '#')
	    return next_statement();

	if (command.equals(tok_else)) {
	    if (!find_statement(tok_endif)) {
		errnum = ERR_ENDIF;
		return false;
	    }
	    return next_statement();
	}

	if (command.equals(tok_endif)) return next_statement();

	if (command.equals(tok_enduntil)) {
	    int linenum = ((Integer) loops.lastElement()).intValue();
	    String condition = (String) (commands.elementAt(linenum).toString());
	    if (test(direction(condition), what(condition))) {
		loops.removeElementAt(loops.size() - 1);
	    } else {
		currcommand = linenum;
		deadcycles += 1;
		if (deadcycles >= maxdeadcycles) {
		    errnum = ERR_DEADCYCLES;
		    return false;
		}
	    }
	    return next_statement();
	}


	if (command.equals(tok_goahead)) {
	    if (!test(AHEAD, Globals.BLOCK)) {
		player.move();
		deadcycles = 0;
		return true;
	    } else {
		errnum = ERR_BUMP;
		return false;
	    }
	} else if (command.equals(tok_turnright)) {
	    player.adjustDirection(RIGHT);
	    deadcycles = 0;
	    return true;
	} else if (command.equals(tok_turnaround)) {
	    player.adjustDirection(BEHIND);
	    deadcycles = 0;
	    return true;
	} else if (command.equals(tok_turnleft)) {
	    player.adjustDirection(LEFT);
	    deadcycles = 0;
	    return true;
	}

	if (command.startsWith(tok_if)) {
	    if (test(direction(command), what(command))) {
		return next_statement();
	    } else {
		if (!find_statement(tok_else)) {
		    errnum = ERR_ENDIF;
		    return false;
		}
		return next_statement();
	    }
	}

	if (command.startsWith(tok_until)) {
	    if (test(direction(command), what(command))) {
		if (!find_statement(tok_enduntil)) {
		    errnum = ERR_ENDUNTIL;
		    return false;
		}
	    } else {
		loops.addElement(new Integer(currcommand));
	    }
	    return next_statement();
	}

	errnum = ERR_SYNTAX;
	return false;
    }
}
